#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   fistr2abaqus.py
#
#       fistr から abaqus 形式にメッシュ変換
#
#   18/01/28    新規作成
#

import os
import sys

ls = "\n"

fistrName = ""

#要素名とnodeNo
elOrder = [
           # fistr   nodeNo(abaqus)       abaqus
            ["341", [1, 2, 3, 4],         "C3D4"],
            ["361", [1,2,3,4,5,6,7,8],    "C3D6"],
            ["342", [1,2,3,4,6,7,5,8,9,10], "C3D10"],
            ["362", [1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20], "C3D20"]
          ]

#要素とfaceNo
faceOrder = [
           # fistr   面No
            ["341", [1,2,3,4]],
            ["361", [1,2,3,4,5,6]],
            ["342", [1,2,3,4]],
            ["362", [1,2,3,4,5,6]]
            ]

#  deleteSp
#    空白を削除
def deleteSp(line):
    newLine = ""
    for chara in line:
        if chara != " " and chara != "\n" and chara != "\r":
            newLine += chara
    return newLine

#  getValName
#    valName(TYPEなど）を取得
def getValName(line, valName):
    valName = valName.upper()
    words = line.split(",")
    name = ""
    for word in words:
        # 「TYPE=」などを検索
        if word.upper().find(valName+"=") >= 0:
            name = word.split("=")[1]
            break
    newName = ""
    for chara in name:
        if chara > " ":
            newName += chara
    return newName

#  setValName
#    変数の値をセット
#    行末コードにより処理を変更（importして使う事がある為）
#    　abaqus2fistrの中のみなら問題なし
def setValName(line, val):
    valNames = val.split("=")
    words = deleteSp(line).split(",")
    i = 0
    flag = 0
    for word in words:
        name = (word.split()[0]).split("=")[0].upper()
        if name == valNames[0]:
            words[i] = " " + val
            flag = 1
            break
        i += 1
    if flag == 0:
        words.append(" "+val)
    newLine = ",".join(words) + ls
    return newLine

#
#  getNodeLines
#    nodeの行を取得
def getNodeLines(headerConts):
    newLines = []
    for header in headerConts:
        words = deleteSp(header[0]).split(",")
        if words[0] == "!NODE":
            line = "*NODE" + header[0][len("!NODE"):]
            newLines.append(line)
            newLines += header[1]
    return newLines

#
#  changeElementOrder
#    nodeNoを要素type毎に並べ替える
def changeElementOrder(elType, dataLines):
    orders = []
    abaType = ""
    for el in elOrder:
        if el[0] == elType:
            orders = el[1]
            abaType = el[2]
            break
    newLines = []
    for line in dataLines:
        words = deleteSp(line).split(",")
        newWords = [""]*len(words)
        newWords[0] = words[0]
        i = 1
        for idx in orders:
            nodeNo = words[idx]
            newWords[i] = nodeNo
            i += 1
        newLine = ", ".join(newWords) + ls
        newLines.append(newLine)
    header = ["*ELEMENT, TYPE=" + abaType + ls]
    newLines = header + newLines
    return newLines

#
#  getElementLines
#    elementの行を取得
def getElementLines(headerConts):
    newLines = []
    for header in headerConts:
        words = deleteSp(header[0]).split(",")
        if words[0] == "!ELEMENT":
            line = "*ELEMENT" + header[0][len("!ELEMENT"):]
            elType = getValName(line, "TYPE")
            elGrp = getValName(line, "EGRP")
            elLines = changeElementOrder(elType, header[1])
            if elGrp != "":
                elLines[0] = setValName(elLines[0], "ELSET="+elGrp)
            newLines += elLines
    return newLines

#
#  getSuGrpLines
#    面group行を取得（Sを追加するのみ）
def getSuGrpLines(headerConts):
    newLines = []
    for header in headerConts:
        words = deleteSp(header[0]).split(",")
        if words[0] == "!SGROUP":
            grpName = getValName(header[0], "SGRP")
            line = "*SURFACE, NAME=" + grpName + ls
            newLines.append(line)
            for faceLine in header[1]:
                words = deleteSp(faceLine).split(",")
                #面Noに「S」を追加するのみ
                words[1] = "S" + words[1]
                line = ", ".join(words) + ls
                newLines.append(line)
    return newLines

#
#  getNdGrpLines
#    nodeGroup行を取得
def getNdGrpLines(headerConts):
    newLines = []
    for header in headerConts:
        words = deleteSp(header[0]).split(",")
        if words[0] == "!NGROUP":
            grpName = getValName(header[0], "NGRP")
            line = "*NSET, NSET=" + grpName + ls
            newLines.append(line)
            newLines += header[1]
    return newLines

#
#  getHeaderConts
#    header毎にデータ（行）を取得
def getHeaderConts(lines):
    headerConts = []
    header = ["", []]
    for line in lines:
        if line[0] == "!":
            if header[0] != "":
                headerConts.append(header)
            header = [line, []]
        else:
            header[1].append(line)
    if header[0] != "":
        headerConts.append(header)
    return headerConts

#
#  fistr2abaqus
#  ------------
#    fistrからabaqus形式に変換
def fistr2abaqus(fileDir, fName):
    print
    print "fistr mesh will be converted to abaqus."
    fileName = fileDir + os.sep + fName + ".msh"
    f = open(fileName)
    lines = f.readlines()
    f.close()
    headerConts = getHeaderConts(lines)
    nodeLines = getNodeLines(headerConts)
    elLines = getElementLines(headerConts)
    suGrpLines = getSuGrpLines(headerConts)
    ndGrpLines = getNdGrpLines(headerConts)
    newLines = nodeLines + elLines + suGrpLines + ndGrpLines
    abaFileName = fileDir + os.sep + fName + ".inp"
    f = open(abaFileName, "w")
    for line in newLines:
        f.write(line)
    f.close()
    print "  converted.  done !"


if __name__ == "__main__":
    fistrDir = sys.argv[1]
    fileDir = os.sep.join(fistrDir.split(os.sep)[:-1])
    if fileDir == "":
        fileDir = os.getcwd()
    fistrName = fistrDir.split(os.sep)[-1]
    if fistrName.split(".")[-1] == "msh":
        name = fistrName[:-4]
    else:
        name = fistrName
    fistr2abaqus(fileDir, name)

