#!/usr/bin/env python  
# -*- coding: utf-8 -*-    
#
#   frd2vtk.py
#
#       calculixのfrdファイルをvtk形式に変換する
#
#   18/07/07    新規作成
#      08/04    tetra2次、hexa2次要素を追加
#


import os
import sys
import glob
import math

nodeLoc = []

def helpMsg():
    cont = u"\n"
    cont += u"-------- frd2vtk.py ---------------------------------\n"
    cont += u"calculix用のfrdファイルを使って、vtkファイルに変換する。\n"
    cont += u"frdファイルからから応力、変位を取得する。\n"
    cont += u"対応している要素は、Tetra1次2次、Hexa1次2次要素のみ。\n"
    cont += u"\n"
    cont += u"<使い方>\n"
    cont += u"inpres2vtk.py <resFile> [<vtkFile>]\n"
    cont += u"    resFile:  frdFile\n"
    cont += u"    vtkFile:  vtkFile （省略時は、resFileの拡張子をvtkに変えて設定される）\n"
    cont += u"\n"
    cont += u"<使用例>\n"
    cont += u"inpres2vtk.py fe\n"
    cont += u"inpres2vtk.py fe.frd test.vtk\n"
    cont += u"\n"
    print cont

def readFileLines(fileName):
    f = open(fileName)
    lines = f.readlines()
    f.close()
    return lines
    
def writeFileLines(fileName, lines):
    f = open(fileName, "w")
    for line in lines:
        f.write(line)
    f.close()

def deleteSp(line):
    newLine = ""
    for chara in line:
        if chara > " ":
            newLine += chara
    return newLine

def getNodesFromInpFile(lines):
    nodes = []
    flag = 0
    nNpdes = 0
    for line in lines:
        words = line.split()
        if len(words) > 0:
            if words[0] == "2C":
                nNodes = int(words[1])
                flag = 1
            else:
                if flag == 1:
                    nodeNo = int(line[3:13])
                    x = float(line[13:25])
                    y = float(line[25:37])
                    z = float(line[37:49])
                    nodes.append([nodeNo, x, y, z])
                    nNodes -= 1
                    if nNodes <= 0:
                        break
    return nodes

def getElementsFromInpFile(lines):
    elements = []
    flag = 0
    nEl = 0
    elNo = 1
    for line in lines:
        words = line.split()
        if len(words) > 0:
            if words[0] == "3C":
                nEl = int(words[1])
                flag = 1
                data = []
            else:
                if flag == 1:
                    if words[0] == "-1":
                        if len(data) > 1:
                            elements.append(data)
                        elNo = int(words[1])
                        data = [elNo]
                    elif words[0] == "-2":
                        for word in words[1:]:
                            data.append(int(word))
                    elif words[0] == "-3":
                        elements.append(data)
                        flag = 0
                        break
                """
                if flag == 1 and words[0] == "-2":
                    words = line.split()
                    data = [elNo]
                    for word in words[1:]:
                        data.append(int(word))
                    elements.append(data)
                    nEl -= 1
                    elNo += 1
                    if nEl <= 0:
                        break
                """
    return elements

def renumberNodeElementNo(nodes, elements):
    global nodeLoc
    nodes.sort()
    maxNodeNo = nodes[-1][0]
    #配列を準備
    nodeLoc = []
    for i in range(maxNodeNo+1):
        nodeLoc.append(["", "", ""])
    #座標を保存
    for node in nodes:
        nodeNo = node[0]
        nodeLoc[nodeNo] = [node[1:], "", ""]
    #要素で使っているnodeNoを設定
    for element in elements:
        for nodeNo in element[1:]:
            nodeLoc[nodeNo][1] = 1
    #要素で使っているnodeのみを取得
    newNodes = []
    newNodeNo = 0
    for i in range(maxNodeNo+1):
        if nodeLoc[i][1] == 1:
            node = [i] + nodeLoc[i][0]
            newNodes.append(node)
            nodeLoc[i][2] = newNodeNo
            newNodeNo += 1
    #要素で使っているnodeNoを更新
    newElements = []
    for i in range(len(elements)):
        elNo = elements[i][0]
        nodeData = []
        for nodeNo in elements[i][1:]:
            newNodeNo = nodeLoc[nodeNo][2]
            nodeData.append(newNodeNo)
        element = [elNo] + nodeData
        newElements.append(element)
    return [newNodes, newElements]

def getDisplacementFromDatFile(lines):
    dispData = []
    ist = len(lines)
    for i in range(len(lines)):
        line = lines[i]
        if line.find("displacements") >= 0:
            ist = i + 1
            break
    i = ist
    while i < len(lines):
        line = lines[i]
        words = line.split()
        if len(words) == 4:
            try:
                dx = float(words[1])
                dy = float(words[2])
                dz = float(words[3])
                dispData.append([dx, dy, dz])
            except:
                break
        i += 1
    return dispData
    
def getDisplacementFromResFile(lines, i):
    dispData = []
    ist = 0
    #i = len(lines) -1
    while i < len(lines):
        line = lines[i]
        words = line.split()
        if len(words) > 1:
            if words[1] == "DISP":
                ist = i
                break
        i += 1
    if ist == 0:
        return [dispData, i]

    i = ist + 5
    while i < len(lines):
        line = lines[i]
        words = line.split()
        if int(words[0]) == -1:
            nodeNo = int(line[3:13])
            dx = float(line[13:25])
            dy = float(line[25:37])
            dz = float(line[37:49])
            dispData.append([dx, dy, dz])
        else:
            break
        i += 1
    return [dispData, i]

def getTensorFromResFile(lines, i):
    tensorData = []
    ist = 0
    #i = ist - 1
    while i < len(lines):
        line = lines[i]
        words = line.split()
        if len(words) > 1:
            if words[1] == "STRESS":
                ist = i
                break
        i += 1
    if ist == 0:
        return [tensorData, i]

    i = ist + 7
    while i < len(lines):
        line = lines[i]
        words = line.split()
        if int(words[0]) == -1:
            nodeNo = int(line[3:13])
            xx = float(line[13:25])
            yy = float(line[25:37])
            zz = float(line[37:49])
            xy = float(line[49:61])
            yz = float(line[61:73])
            zx = float(line[73:85])
            tensorData.append([xx,yy,zz,xy,yz,zx])
        else:
            break
        i += 1
    return [tensorData, i]

def getMisesStressFromTensorData(tensorData):
    misesStress = []
    for tensor in tensorData:
        [xx, yy, zz, xy, yz, zx] = tensor
        a = 0.5*(math.pow((xx-yy),2) + math.pow((yy-zz),2) + math.pow((zz-xx),2)) + 3.0*(xy*xy + yz*yz + zx*zx)
        mises = math.pow(a, 0.5)
        misesStress.append([mises])
    return misesStress

def createVtkHeader():
    lines = ["# vtk DataFile Version 2.0\n"]
    lines += ["surface\n"]
    lines += ["ASCII\n"]
    lines += ["DATASET UNSTRUCTURED_GRID\n"]
    return lines

def createVtkNodes(nodes):
    lines = ["POINTS " + str(len(nodes)) + " float\n"]
    for node in nodes:
        strData = []
        for xyz in node[1:]:
            strData.append(str(xyz))
        line = " ".join(strData) + "\n"
        lines += [line]
    return lines

#
#  createVtkElements
#    要素行を作成（三角形1次要素のみ）
def createVtkElements(elements):
    nElement = len(elements)
    nAllData = nElement * 7
    lines = ["CELLS " + str(nElement) + " " + str(nAllData) + "\n"]
    for element in elements:
        strData = ["6"]
        for nodeNo in element[1:]:
            strData.append(str(nodeNo))
        line = " ".join(strData) + "\n"
        lines += [line]
    lines += ["CELL_TYPES " + str(nElement) + "\n"]
    for element in elements:
        lines += ["13\n"]
    return lines
    
#
#  createVtkElementsTetra
#    要素行を作成（四面体1次）
def createVtkElementsTetra(elements):
    nElement = len(elements)
    nAllData = nElement * 5
    lines = ["CELLS " + str(nElement) + " " + str(nAllData) + "\n"]
    for element in elements:
        strData = ["4"]
        for nodeNo in element[1:]:
            strData.append(str(nodeNo))
        line = " ".join(strData) + "\n"
        lines += [line]
    lines += ["CELL_TYPES " + str(nElement) + "\n"]
    for element in elements:
        lines += ["10\n"]
    return lines

#
#  createVtkElementsHexa
#    要素行を作成（六面体）
def createVtkElementsHexa(elements):
    nElement = len(elements)
    nAllData = nElement * 9
    lines = ["CELLS " + str(nElement) + " " + str(nAllData) + "\n"]
    for element in elements:
        strData = ["8"]
        for nodeNo in element[1:]:
            strData.append(str(nodeNo))
        line = " ".join(strData) + "\n"
        lines += [line]
    lines += ["CELL_TYPES " + str(nElement) + "\n"]
    for element in elements:
        lines += ["12\n"]
    return lines

#
#  createVtkElementsTetra2nd(elements)
#    要素行を作成（四面体2次）
def createVtkElementsTetra2nd(elements):
    nElement = len(elements)
    nAllData = nElement * 11
    lines = ["CELLS " + str(nElement) + " " + str(nAllData) + "\n"]
    for element in elements:
        strData = ["10"]
        for nodeNo in element[1:]:
            strData.append(str(nodeNo))
        line = " ".join(strData) + "\n"
        lines += [line]
    lines += ["CELL_TYPES " + str(nElement) + "\n"]
    for element in elements:
        lines += ["24\n"]
    return lines

#
#  createVtkElementsHexa2nd(elements)
#    要素行を作成（六面体2次）
def createVtkElementsHexa2nd(elements):
    nElement = len(elements)
    nAllData = nElement * 21
    lines = ["CELLS " + str(nElement) + " " + str(nAllData) + "\n"]
    for element in elements:
        strData = ["20"]
        for nodeNo in element[1:]:
            strData.append(str(nodeNo))
        line = " ".join(strData) + "\n"
        lines += [line]
    lines += ["CELL_TYPES " + str(nElement) + "\n"]
    for element in elements:
        lines += ["25\n"]
    return lines


def reorderElementsHex2nd(elements):
    newElements = []
    for elm in elements:
        newElement = elm[:13] + elm[17:] + elm[13:17]
        newElements.append(newElement)
    return newElements
    
#
#  createVtkNodeData
#    nodeデータのfieldを作成
def createVtkNodeData(dataList):
    nData = len(dataList[0][2])
    lines = ["POINT_DATA " + str(nData) + "\n"]
    lines += ["FIELD attributes " + str(len(dataList)) + "\n"]
    for data in dataList:
        name = data[0]
        dataType = data[1]
        vals = data[2]
        strData = [name, str(len(vals[0])), str(nData), dataType]
        lines += [" ".join(strData) + "\n"]
        for val in vals:
            strData = []
            for a in val:
                strData.append(str(a))
            line = " ".join(strData) + "\n"
            lines += [line]
    return lines
    
#
#  createVtkCellData(dataList)
#    cellデータのfieldを作成
def createVtkCellData(dataList):
    nData = len(dataList[0][2])
    lines = ["CELL_DATA " + str(nData) + "\n"]
    lines += ["FIELD attributes " + str(len(dataList)) + "\n"]
    for data in dataList:
        name = data[0]
        dataType = data[1]
        vals = data[2]
        strData = [name, str(len(vals[0])), str(nData), dataType]
        lines += [" ".join(strData) + "\n"]
        for val in vals:
            strData = []
            for a in val:
                strData.append(str(a))
            lines += [" ".join(strData) + "\n"]
    return lines

def convFrd2vtk(inpresFile, vtkFile):
    inpLines = readFileLines(inpresFile)
    nodes = getNodesFromInpFile(inpLines)
    elements = getElementsFromInpFile(inpLines)
    if len(elements[0]) == 9:
        #hexa 六面体
        [nodes, elements] = renumberNodeElementNo(nodes, elements)
        #vtk作成
        vtkLines = createVtkHeader()
        vtkLines += createVtkNodes(nodes)
        vtkLines += createVtkElementsHexa(elements)
    elif len(elements[0]) == 5:
        #tetra 四面体
        [nodes, elements] = renumberNodeElementNo(nodes, elements)
        #vtk作成
        vtkLines = createVtkHeader()
        vtkLines += createVtkNodes(nodes)
        vtkLines += createVtkElementsTetra(elements)
    elif len(elements[0]) == 6:
        #tri 三角形shell
        [nodes, elements] = renumberNodeElementNo(nodes, elements)
        #節点、要素のvtk作成
        vtkLines = createVtkHeader()
        vtkLines += createVtkNodes(nodes)
        vtkLines += createVtkElements(elements)
    elif len(elements[0]) == 21:
        #hexa2nd 六面体2次
        [nodes, elements] = renumberNodeElementNo(nodes, elements)
        elements = reorderElementsHex2nd(elements)  #要素の並び順を変更
        #vtk作成
        vtkLines = createVtkHeader()
        vtkLines += createVtkNodes(nodes)
        vtkLines += createVtkElementsHexa2nd(elements)
    elif len(elements[0]) == 11:
        #tetra2nd 四面体2次
        [nodes, elements] = renumberNodeElementNo(nodes, elements)
        #vtk作成
        vtkLines = createVtkHeader()
        vtkLines += createVtkNodes(nodes)
        vtkLines += createVtkElementsTetra2nd(elements)
    else:
        print "要素タイプには、対応していません。"
        exit()
    
    resFile = ".".join(inpresFile.split(".")[:-1]) + ".frd"
    if len(glob.glob(resFile)) != 0:
        dataList = []
        resLines = readFileLines(resFile)
        iDisp = 0; iTens = 0; i = 0
        while (iDisp < len(resLines)) and (iTens < len(resLines)):
            dataList = []
            [dispData, iDisp] = getDisplacementFromResFile(resLines, iDisp)
            if len(dispData) != 0:
                dataList.append(["DISPLACEMENT", "float", dispData])
            [tensorData, iTens] = getTensorFromResFile(resLines, iTens)
            if len(tensorData) != 0:
                dataList.append(["STRESS", "float", tensorData])
                misesData = getMisesStressFromTensorData(tensorData)
                dataList.append(["misesStress", "float", misesData])
            if len(dataList) != 0:
                dataLines = createVtkNodeData(dataList)
                allLines = vtkLines + dataLines
                fileName = ".".join(vtkFile.split(".")[:-1]) + "." + str(i) + ".vtk"
                print "creating " + fileName + "..."
                writeFileLines(fileName, allLines)
            i += 1

if __name__ == "__main__":
    if len(sys.argv) == 2:
        inpresFile = sys.argv[1]
        vtkFile = ""
    elif len(sys.argv) == 3:
        inpresFile = sys.argv[1]
        vtkFile = sys.argv[2]
    else:
        print "\nERROR:引数が合いません。"
        helpMsg()
        exit()
    if inpresFile == "-h" or inpresFile == "-help" or inpresFile == "--help":
        helpMsg()
        exit()

    if inpresFile.split(".")[-1] != "frd":
        inpresFile += ".frd"
    if vtkFile == "":
        vtkFile = ".".join(inpresFile.split(".")[:-1]) + ".vtk"
    elif vtkFile.split(".")[-1] != "vtk":
        vtkFile += ".vtk"

    convFrd2vtk(inpresFile, vtkFile)
    print u"vtk変換しました。"
    
    
